"""
ROS1 version
"""

import rospy
import rostopic

import armw.globals
from armw.base_node import BaseNode


class Ros1Node(BaseNode):
    def __init__(self):
        super().__init__()

    def spin(self):
        # Issue status periodically.
        rate = armw.Rate(self.update_rate)
        while not armw.is_shutdown() and not self.stop_commanded:
            self.loop_once()

            try:
                rate.sleep()
            except rospy.ROSInterruptException:
                self.log_warn(f"{armw.globals.NODE_NAME} Shutting down")

        self.shutdown()

    def log(self, level: int, message: str):
        if level == -1:
            rospy.logdebug(message)
        elif level == 0:
            rospy.loginfo(message)
        elif level == 1:
            rospy.logwarn(message)
        elif level == 2:
            rospy.logerr(message)
        elif level == 3:
            rospy.logfatal(message)

    def log_throttle(self, interval: int, level: int, message: str):
        if level == -1:
            rospy.logdebug_throttle(interval, message)
        elif level == 0:
            rospy.loginfo_throttle(interval, message)
        elif level == 1:
            rospy.logwarn_throttle(interval, message)
        elif level == 2:
            rospy.logerr_throttle(interval, message)

    def get_name(self):
        return rospy.get_name()

    def get_namespace(self):
        return rospy.get_namespace()

    def search_param(self, parameter_name):
        return rospy.search_param(parameter_name)

    def has_param(self, parameter_name):
        return rospy.has_param(parameter_name)

    def get_param(self, parameter_name, default=None):
        """
        Returns the value for a given parameter name, or None if it can't find it
        """
        return rospy.get_param(parameter_name, default)

    def set_param(self, parameter_name, value):
        rospy.set_param(parameter_name, value)

    def delete_param(self, parameter_name):
        rospy.delete_param(parameter_name)

    def get_topic_type(self, topic):
        topic_type, _, _ = rostopic.get_topic_class(topic)
        return topic_type

    def publish(self, topic: str, message_type, queue_size=1, latch=False):
        return rospy.Publisher(topic, message_type, queue_size=queue_size, latch=latch)

    def subscribe(self, topic: str, message_type, callback, queue_size=1):
        if message_type is None:
            from rospy import AnyMsg
            message_type = AnyMsg

        return rospy.Subscriber(topic, message_type, callback)

    def create_service_server(self, name, service_type, callback):
        return rospy.Service(name, service_type, callback)

    def create_service_client(self, name: str, service_type, persistent=False):
        return rospy.ServiceProxy(name, service_type, persistent=persistent)

    def wait_for_message(self, topic, message_type, timeout=None):
        if message_type is None:
            from rospy import AnyMsg
            message_type = AnyMsg

        if timeout is None:
            return rospy.wait_for_message(topic, message_type)
        else:
            return rospy.wait_for_message(topic, message_type, timeout=timeout)

    def wait_for_service(self, name, timeout=None):
        if timeout is None:
            rospy.wait_for_service(name)
        else:
            return rospy.wait_for_service(name, timeout=timeout)


def init_node(node_name):
    rospy.init_node(node_name)


def run_node(node_object):
    armw.globals.NODE = node_object
    try:
        armw.globals.NODE.spin()
        rospy.loginfo(f"{armw.globals.NODE_NAME} shutdown")
    except rospy.ROSInterruptException:
        rospy.loginfo(f"{armw.globals.NODE_NAME} shutdown (interrupt)")


def main(node_class: type(BaseNode), node_name: str):
    rospy.init_node(node_name)

    armw.globals.NODE_NAME = node_name
    armw.globals.NODE = node_class()

    try:
        armw.globals.NODE.spin()
        rospy.loginfo(f"{node_name} shutdown")
    except rospy.ROSInterruptException:
        rospy.loginfo(f"{node_name} shutdown (interrupt)")



# Straight up copied from the ros2 rolling distro

from typing import Union

from rclpy.impl.implementation_singleton import rclpy_implementation as _rclpy
from rclpy.node import Node
from rclpy.qos import QoSProfile
from rclpy.signals import SignalHandlerGuardCondition
from rclpy.utilities import timeout_sec_to_nsec


def wait_for_message(msg_type, node: 'Node', topic: str, *, qos_profile: Union[QoSProfile, int] = 1, time_to_wait=-1):
    """
    Wait for the next incoming message.

    :param msg_type: message type
    :param node: node to initialize the subscription on
    :param topic: topic name to wait for message
    :param qos_profile: QoS profile to use for the subscription
    :param time_to_wait: seconds to wait before returning
    :returns: (True, msg) if a message was successfully received, (False, None) if message
        could not be obtained or shutdown was triggered asynchronously on the context.
    """
    context = node.context
    wait_set = _rclpy.WaitSet(1, 1, 0, 0, 0, 0, context.handle)
    wait_set.clear_entities()

    sub = node.create_subscription(msg_type, topic, lambda _: None, qos_profile=qos_profile)
    try:
        wait_set.add_subscription(sub.handle)
        sigint_gc = SignalHandlerGuardCondition(context=context)
        wait_set.add_guard_condition(sigint_gc.handle)

        timeout_nsec = timeout_sec_to_nsec(time_to_wait)
        wait_set.wait(timeout_nsec)

        subs_ready = wait_set.get_ready_entities('subscription')
        guards_ready = wait_set.get_ready_entities('guard_condition')

        if guards_ready:
            if sigint_gc.handle.pointer in guards_ready:
                return False, None

        if subs_ready:
            if sub.handle.pointer in subs_ready:
                msg_info = sub.handle.take_message(sub.msg_type, sub.raw)
                if msg_info is not None:
                    return True, msg_info[0]
    finally:
        node.destroy_subscription(sub)

    return False, None



"""
Base class to provide a reasonable API to set up runtime-reconfigurable parameters
"""

from armw.base_node import BaseNode


class BaseParameterServer(object):
    def __init__(self, node_object: BaseNode, namespace: str = ""):
        self.node = node_object
        self.namespace = namespace

    def add_parameter(self, name, description, default_value, min_value=None, max_value=None, parameter_type=None):
        raise NotImplementedError

    def start(self, callback):
        raise NotImplementedError






#!/usr/bin/env python

"""
Dynamic dynamic reconfigure server.

Just register your variables for the dynamic reconfigure
and call start with a callback.

Author: Sammy Pfeiffer

BSD 3-Clause License

Copyright (c) 2020, PAL Robotics S.L.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from dynamic_reconfigure.server import Server
from dynamic_reconfigure.parameter_generator_catkin import ParameterGenerator
from dynamic_reconfigure.encoding import extract_params
from rospkg import RosPack
import rospy


class DDynamicReconfigure(ParameterGenerator):
    """Dynamic reconfigure server that can be instanced directly."""

    def __init__(self, name=None):
        global id
        self.group = self.Group(self, "Default", "", True, 0, 0)
        id = 1
        if name is None:
            self.name = rospy.get_name() + "_dyn_rec"
        else:
            self.name = name
        self.constants = []
        rp = RosPack()
        self.dynconfpath = rp.get_path('dynamic_reconfigure')

    def get_type(self):
        class TypeClass(object):
            def __init__(self, config_description):
                self.config_description = config_description
                self.min = {}
                self.max = {}
                self.defaults = {}
                self.level = {}
                self.type = {}
                self.all_level = 0

                for param in extract_params(config_description):
                    self.min[param['name']] = param['min']
                    self.max[param['name']] = param['max']
                    self.defaults[param['name']] = param['default']
                    self.level[param['name']] = param['level']
                    self.type[param['name']] = param['type']
                    self.all_level = self.all_level | param['level']
        return TypeClass(self.group.to_dict())

    def add_variable(self, name, description, default=None, min=None, max=None, edit_method=""):
        """Register variable, like gen.add() but deducting the type"""
        if type(default) == int:
            if edit_method == "":
                self.add(name, "int", 0, description, default, min, max)
            else: # enum
                self.add(name, "int", 0, description, default, min, max, edit_method)
        elif type(default) == float:
            self.add(name, "double", 0, description, default, min, max)
        elif type(default) == str:
            self.add(name, "str", 0, description, default)
        elif type(default) == bool:
            self.add(name, "bool", 0, description, default)

        return default


    def get_variable_names(self):
        """Return the names of the dynamic reconfigure variables"""
        names = []
        for param in self.group.parameters:
            names.append(param['name'])
        return names


    def start(self, callback):
        self.dyn_rec_srv = Server(self.get_type(), callback, namespace=self.name)


"""
Functions to handle importing interface definitions
"""

import importlib
from typing import List, Dict


def import_message(package_name: str, message_name: str):
    if package_name == "std_msgs" and message_name == "Time":
        # ROS changing the name of this causes SO MANY problems
        package_name = "builtin_interfaces"
        message_name = "Time"
    elif package_name == "rosgraph_msgs" and message_name == "Log":
        package_name = "rcl_interfaces"
        message_name = "Log"

    return getattr(importlib.import_module(f"{package_name}.msg"), message_name)


def import_service_(package_name: str, service_name: str):
    service_class = getattr(importlib.import_module(f"{package_name}.srv"), service_name)
    return service_class, service_class.Response, service_class.Request


def deserialze_anymsg(msg_data):
    return msg_data


def get_msg_type_from_object(msg):
    string = str(type(msg))
    useful_bit = string.split("'")[1]
    pkg = useful_bit.split(".")[0]
    msg_name = useful_bit.split(".")[-1].split("_")[-1]
    return f"{pkg}/{msg_name}"


def get_msg_fields_and_types(msg) -> Dict[str, str]:
    return msg.get_fields_and_field_types()


def get_msg_fields(msg) -> List[str]:
    return list(get_msg_fields_and_types(msg).keys())


if __name__ == '__main__':
    print(get_msg_type_from_object(None))



"""
ROS2 version
"""

import time

import armw.globals
import armw.ros2.wait_for_message
import rclpy
from armw.base_node import BaseNode
from rclpy.callback_groups import MutuallyExclusiveCallbackGroup, ReentrantCallbackGroup
from rclpy.node import Node
from rclpy.qos import QoSProfile, QoSDurabilityPolicy
from ros2topic.api import get_msg_class

executor: rclpy.executors.MultiThreadedExecutor = None


class ServiceProxy(object):
    def __init__(self, service_name, service_type, callback_group):
        self._client = armw.NODE().create_client(service_type, service_name, callback_group=callback_group)
        self.service_class = service_type.Request

    # This causes a crash, so I'm going to leave it off
    # def __del__(self):
    #     armw.NODE().destroy_client(self._client)

    def __call__(self, *args, **kwargs):
        if len(args) > 0:
            req = args[0]
        else:
            req = self.service_class(**kwargs)

        future = self._client.call_async(req)
        executor.spin_until_future_complete(future, timeout_sec=1)
        return future.result()


class Ros2Node(BaseNode, Node):
    def __init__(self):
        # Run parent constructors
        Node.__init__(self, armw.globals.NODE_NAME, allow_undeclared_parameters=True, automatically_declare_parameters_from_overrides=True)
        BaseNode.__init__(self)

        self.timer_callback_group = MutuallyExclusiveCallbackGroup()
        self.service_callback_group = ReentrantCallbackGroup()

        # Need so do some silly stuff to make services work
        self.service_callbacks = {}

    def spin(self):
        timer_period = 1.0 / self.update_rate
        self.timer = self.create_timer(timer_period, self.loop_once, callback_group=self.timer_callback_group)

    def log(self, level: int, message: str):
        if level == -1:
            self.get_logger().debug(message)
        elif level == 0:
            self.get_logger().info(message)
        elif level == 1:
            self.get_logger().warn(message)
        elif level == 2:
            self.get_logger().error(message)
        elif level == 3:
            self.get_logger().fatal(message)

    def log_throttle(self, interval: int, level: int, message: str):
        if level == -1:
            self.get_logger().debug(message, throttle_duration_sec=interval)
        elif level == 0:
            self.get_logger().info(message, throttle_duration_sec=interval)
        elif level == 1:
            self.get_logger().warn(message, throttle_duration_sec=interval)
        elif level == 2:
            self.get_logger().error(message, throttle_duration_sec=interval)
        elif level == 3:
            self.get_logger().fatal(message, throttle_duration_sec=interval)

    def get_name(self):
        return Node.get_name(self)

    def get_namespace(self):
        return Node.get_namespace(self)

    def search_param(self, parameter_name):
        """
        This isn't really a thing in ROS2
        """

        return parameter_name

    def has_param(self, parameter_name):
        if parameter_name[0] == "~":
            parameter_name = parameter_name[1:]
        parameter_name = parameter_name.replace("/", ".")

        return self.has_parameter(parameter_name)

    def get_param(self, parameter_name, default=None):
        """
        Returns the value for a given parameter name, or None if it can't find it
        """

        if len(parameter_name) == 0:
            return default

        if parameter_name[0] == "~":
            parameter_name = parameter_name[1:]

        if len(parameter_name) == 0:
            return default

        parameter_name = parameter_name.replace("/", ".")
        if parameter_name[0] == '.':
            parameter_name = parameter_name[1:]

        try:
            # TODO: Do we want to support reading in dictionary parameters if the default isn't a dictionary?
            if default is not None and isinstance(default, dict):
                return self.get_param_as_dict(parameter_name, default)
            else:
                param_value = self.get_parameter(parameter_name).get_parameter_value()

                if param_value.type == 0:  # Parameter not set
                    return default
                return rclpy.parameter.parameter_value_to_python(param_value)
        except Exception as e:
            self.log_error(f"Unable to read parameter {parameter_name}: {e}")
            return default

    def get_param_as_dict(self, parameter_name, default=None):
        param_dict = self.get_parameters_by_prefix(parameter_name)

        if len(list(param_dict.keys())) == 0:
            return default

        return_value = {}
        for key in param_dict:
            param_value = param_dict[key].get_parameter_value()
            raw_value = rclpy.parameter.parameter_value_to_python(param_value)
            return_value[key] = raw_value

        return return_value

    def set_param(self, parameter_name, value):
        if parameter_name[0] == "~":
            parameter_name = parameter_name[1:]
        parameter_name = parameter_name.replace("/", ".")

        parameter = rclpy.Parameter(parameter_name, rclpy.Parameter.Type.from_parameter_value(value), value)
        self.set_parameters([parameter])

    def fix_topic_name(self, topic):
        if topic[0] == "~":
            topic = self.get_name() + topic[1:]

        return topic

    def get_topic_type(self, topic):
        return get_msg_class(self, topic, blocking=False)

    def publish(self, topic: str, message_type, queue_size=1, latch=False):
        if latch:
            qos = QoSProfile(depth=1, durability=QoSDurabilityPolicy.TRANSIENT_LOCAL)
        else:
            qos = queue_size

        return self.create_publisher(message_type, self.fix_topic_name(topic), qos)

    def subscribe(self, topic: str, message_type, callback, queue_size=1):
        # This will work if there's a publisher up, but it's not recommended
        if message_type is None:
            self.log_warn("-" * 1000)
            self.log_warn(f"Subscribing to anymsg on topic {topic}.  In ROS2 this requires waiting for a publisher to exist")
            message_type = get_msg_class(self, topic, blocking=True)

        return self.create_subscription(message_type, self.fix_topic_name(topic), callback, queue_size)

    def create_service_server(self, name, service_type, callback):
        self.service_callbacks[name] = callback

        return self.create_service(service_type, name, lambda req, resp: self.service_callback(req, resp, name))

    def service_callback(self, request, response, name):
        response = self.service_callbacks[name](request)
        return response

    def create_service_client(self, name: str, service_type, persistent=False):
        # return self.create_client(service_type, name)
        return ServiceProxy(name, service_type, self.service_callback_group)

    def wait_for_message(self, topic_name, topic_type, timeout=None):
        if timeout is None:
            time_to_wait = 100000  # Essentially forever
        else:
            time_to_wait = timeout

        self.log_warn(f"Waiting for message on {topic_name} for {time_to_wait} seconds")

        if topic_type is None:
            from ros2topic.api import get_msg_class
            topic_type = get_msg_class(self, topic_name, blocking=True)

        success, msg = armw.ros2.wait_for_message.wait_for_message(topic_type, self, topic_name, time_to_wait=time_to_wait)

        self.log_warn(f"Got message on {topic_name}")
        return msg

    def wait_for_service(self, service_name, timeout=None):
        """
        This might not work correctly yet
        """

        # Resolve the name
        if service_name[0] != "/":
            service_name = self.get_namespace() + "/" + service_name
            service_name = armw.canonicalize_name(service_name)

        # Wait
        self.log_warn(f"Waiting for service {service_name}")
        while True:
            for service in self.get_service_names_and_types():
                if service[0] == service_name:
                    self.log_warn(f"Got service {service_name}")
                    return
            time.sleep(0.1)


def init_node(node_name):
    rclpy.init()
    armw.globals.NODE_NAME = node_name


def run_node(node_object):
    global executor

    armw.globals.NODE = node_object

    executor = rclpy.executors.MultiThreadedExecutor()
    try:
        armw.globals.NODE.spin()
        executor.add_node(armw.globals.NODE)
        executor.spin()
        rclpy.shutdown()
    except KeyboardInterrupt:
        pass

    print(f"{armw.globals.NODE_NAME} shutdown")


def main(node_class, node_name):
    global executor

    rclpy.init()
    executor = rclpy.executors.MultiThreadedExecutor()

    armw.globals.NODE_NAME = node_name
    armw.globals.NODE = node_class()

    try:
        armw.globals.NODE.spin()
        executor.add_node(armw.globals.NODE)
        executor.spin()
        rclpy.shutdown()
    except KeyboardInterrupt:
        pass

    print(f"{node_name} shutdown")



def package_and_msg_name(name):
    """
    https://docs.ros.org/en/diamondback/api/roslib/html/python/roslib.names-pysrc.html#package_resource_name
    """
    name.replace('/msg/', '/')
    name.replace('/srv/', '/')
    if "/" in name:
        val = tuple(name.split("/"))
        if len(val) != 2:
            raise ValueError("invalid name [%s]" % name)
        else:
            return val
    else:
        return '', name


"""
Various functions to explore the ROS world
"""

import rosgraph
import rostopic
import rosnode
import rospkg


def check_master():
    return rosgraph.is_master_online()


def get_topic_list():
    master = rosgraph.Master('/rostopic')
    pubs, _ = rostopic.get_topic_list(master=master)
    return [x[0] for x in pubs]


def get_node_list():
    return rosnode.get_node_names()


def get_package_path(package_name: str):
    return rospkg.RosPack().get_path(package_name)



import inspect

try:
    from ddynamic_reconfigure import DDynamicReconfigure
except Exception as e:
    print(e)

from armw.base_node import BaseNode
from armw.doc_int_enum import DocIntEnum
from armw.base_parameter_server import BaseParameterServer


class Ros1ParameterServer(BaseParameterServer):
    def __init__(self, node_object: BaseNode, namespace: str = ""):
        super().__init__(node_object, namespace)

        self.server = DDynamicReconfigure(namespace)

    def add_parameter(self, name, description, default_value, min_value=None, max_value=None, parameter_type=None):
        edit_method = ""
        if inspect.isclass(parameter_type) and issubclass(parameter_type, DocIntEnum):
            option_list = []
            for value in parameter_type:
                option_list.append(self.server.const(parameter_type.get_field_name(value), "int", int(value), parameter_type.get_description(value)))
            edit_method = self.server.enum(option_list, parameter_type.__name__)

        self.server.add_variable(name, description, default=default_value, min=min_value, max=max_value, edit_method=edit_method)

    def start(self, callback):
        self.server.start(callback)



import types

import rospy
import armw.globals


def get_time():
    if armw.globals.NODE is not None:
        return armw.Time().now().to_sec()
    else:
        raise Exception("time is not initialized. Have you created a node?")


class Duration(rospy.Duration):
    def __init__(self, secs=0, nsecs=0):
        super().__init__(nanoseconds=secs * 1000000000 + nsecs)
        self.to_nsec = types.MethodType(lambda self: self.nanoseconds, self)
        self.to_sec = types.MethodType(lambda self: self.nanoseconds / 1e9, self)
        self.is_zero = types.MethodType(lambda self: self.nanoseconds == 0, self)
        self.secs = secs
        self.nsecs = nsecs

    def __add__(self, other):
        if isinstance(other, Time):
            return armw.Time().from_sec(self.to_sec() + other.to_sec())
        else:
            Duration.__add__(self, other)

    @classmethod
    def from_sec(cls, secs):
        return Duration(int(secs), nsecs=int((secs - int(secs)) * 1000000000))

    @classmethod
    def from_seconds(cls, secs):
        return Duration(int(secs), nsecs=int((secs - int(secs)) * 1000000000))


class Time(rospy.Time):
    def __init__(self, secs=0, nsecs=0):
        super().__init__(secs=secs, nsecs=nsecs)

    def __add__(self, other):
        if isinstance(other, rospy.Duration):
            return armw.Time().from_sec(self.to_sec() + other.to_sec())
        else:
            Time.__add__(self, other)

    def __sub__(self, other):
        if isinstance(other, rospy.Duration):
            return armw.Time().from_sec(self.to_sec() - other.to_sec())
        else:
            Time.__add__(self, other)

    def __gt__(self, other):
        return self.to_sec() > other.to_sec()

    def __lt__(self, other):
        return self.to_sec() < other.to_sec()

    def __ge__(self, other):
        return self.to_sec() >= other.to_sec()

    def __le__(self, other):
        return self.to_sec() <= other.to_sec()

    def __eq__(self, other):
        return self.to_sec() == other.to_sec()

    def __ne__(self, other):
        return self.to_sec() != other.to_sec()

    def to_sec(self):
        return self.sec + float(self.nsecs) / 1000000000

    @classmethod
    def from_sec(cls, secs: float):
        return Time(secs=int(secs), nsecs=int((secs - int(secs)) * 1000000000))

    @classmethod
    def from_seconds(cls, secs: float):
        return Time().from_sec(secs=secs)

    @classmethod
    def now(cls):
        s = rospy.Time.now()
        return Time(secs=s[0], nsecs=s[1])


if __name__ == '__main__':
    a = Time(3, 5)
    b = Duration(5, 6)

    print(a)
    print(b)

    print(a + b)
    print(b + a)



from __future__ import annotations

import armw.globals
import armw.base_node
import armw.base_parameter_server
import armw.doc_int_enum

# Figure out the middleware version

MIDDLEWARE = ""

try:
    import rospy

    MIDDLEWARE = "ros1"
except:
    pass

try:
    import rclpy
    import builtin_interfaces

    MIDDLEWARE = "ros2"
except:
    pass

# Do some stuff with it
if MIDDLEWARE == "ros1":
    import armw.ros_common
    import armw.ros1.ros1_node
    from armw.ros1.ros1_parameter_server import Ros1ParameterServer
    import armw.ros1.interfaces
    import armw.ros1.inspect

    # function and class names
    ArmwNode = armw.ros1.ros1_node.Ros1Node
    ParameterServer = armw.ros1.ros1_parameter_server.Ros1ParameterServer
    Publisher = rospy.Publisher

    Time = rospy.Time
    Duration = rospy.Duration
    Rate = rospy.Rate
    AnyMsg = rospy.AnyMsg
    NativeTime = rospy.Time

    InterruptException = rospy.ROSInterruptException

    main = armw.ros1.ros1_node.main
    init_node = armw.ros1.ros1_node.init_node
    run_node = armw.ros1.ros1_node.run_node
    is_shutdown = rospy.is_shutdown
    sleep = rospy.sleep
    get_time = rospy.get_time
    import_message = armw.ros1.interfaces.import_message
    import_service_ = armw.ros1.interfaces.import_service_
    deserialze_anymsg = armw.ros1.interfaces.deserialze_anymsg
    get_msg_type_from_object = armw.ros1.interfaces.get_msg_type_from_object
    get_msg_fields = armw.ros1.interfaces.get_msg_fields
    get_msg_fields_and_types = armw.ros1.interfaces.get_msg_fields_and_types
    canonicalize_name = rospy.names.canonicalize_name
    resolve_name = rospy.names.resolve_name
    package_and_msg_name = armw.ros_common.package_and_msg_name
    get_package_path = armw.ros1.inspect.get_package_path

    check_master = armw.ros1.inspect.check_master
    get_topic_list = armw.ros1.inspect.get_topic_list
    get_node_list = armw.ros1.inspect.get_node_list

    # Log levels
    DEBUG = rospy.DEBUG
    INFO = rospy.INFO
    WARN = rospy.WARN
    ERROR = rospy.ERROR
elif MIDDLEWARE == "ros2":
    import rclpy.publisher

    import time
    import armw.ros_common
    import armw.ros2.ros2_node
    import armw.ros2.ros2_parameter_server
    import armw.ros2.interfaces
    import armw.ros2.time
    import armw.ros2.util
    import armw.ros2.inspect

    # function and class names
    ArmwNode = armw.ros2.ros2_node.Ros2Node
    ParameterServer = armw.ros2.ros2_parameter_server.Ros2ParameterServer
    Publisher = rclpy.publisher.Publisher

    Time = armw.ros2.time.Time
    Duration = armw.ros2.time.Duration
    Rate = armw.ros2.time.Rate
    AnyMsg = None
    NativeTime = builtin_interfaces.msg.Time

    InterruptException = Exception

    main = armw.ros2.ros2_node.main
    init_node = armw.ros2.ros2_node.init_node
    run_node = armw.ros2.ros2_node.run_node
    is_shutdown = lambda: not rclpy.ok()
    sleep = time.sleep
    get_time = armw.ros2.time.get_time
    import_message = armw.ros2.interfaces.import_message
    import_service_ = armw.ros2.interfaces.import_service_
    deserialze_anymsg = armw.ros2.interfaces.deserialze_anymsg
    get_msg_type_from_object = armw.ros2.interfaces.get_msg_type_from_object
    get_msg_fields = armw.ros2.interfaces.get_msg_fields
    get_msg_fields_and_types = armw.ros2.interfaces.get_msg_fields_and_types
    canonicalize_name = armw.ros2.util.canonicalize_name
    resolve_name = armw.ros2.util.resolve_name
    package_and_msg_name = armw.ros_common.package_and_msg_name
    get_package_path = armw.ros2.inspect.get_package_path

    check_master = armw.ros2.inspect.check_master
    get_topic_list = armw.ros2.inspect.get_topic_list
    get_node_list = armw.ros2.inspect.get_node_list

    # Log levels
    DEBUG = 1
    INFO = 2
    WARN = 3
    ERROR = 4
else:
    print("No middleware available")
    # exit(1)

ParameterEnum = armw.doc_int_enum.DocIntEnum


def import_service(package_name: str, service_name: str):
    if service_name.endswith("Request"):
        service_name = service_name[:-7]
        _, _, req = armw.import_service_(package_name, service_name)
        return req
    elif service_name.endswith("Response"):
        service_name = service_name[:-8]
        _, resp, _ = armw.import_service_(package_name, service_name)
        return resp
    else:
        return armw.import_service_(package_name, service_name)


def import_message_from_name(name):
    pkg, msg = armw.package_and_msg_name(name)
    return armw.import_message(pkg, msg)


def NODE() -> armw.base_node.BaseNode:
    return armw.globals.NODE


def fill_time(dest_object, source_object) -> armw.Time:
    """
    Attempts to get around some time weirdness by doing lots of python weirdness
    """

    seconds_fields = ["sec", "secs", "seconds"]
    nanoseconds_fields = ["nsec", "nsecs", "nanosec", "nanosecs", "nanoseconds"]

    seconds = None
    nanoseconds = None

    for field in seconds_fields:
        if hasattr(source_object, field):
            seconds = getattr(source_object, field)
            break

    for field in nanoseconds_fields:
        if hasattr(source_object, field):
            nanoseconds = getattr(source_object, field)
            break

    if seconds is None:
        raise Exception(f"Source {source_object} does not have field encoding seconds")
    if nanoseconds is None:
        raise Exception(f"Source {source_object} does not have field encoding nanoseconds")

    wrote_seconds = False
    wrote_nanoseconds = False

    for field in seconds_fields:
        if hasattr(dest_object, field):
            setattr(dest_object, field, seconds)
            wrote_seconds = True
            break

    for field in nanoseconds_fields:
        if hasattr(dest_object, field):
            setattr(dest_object, field, nanoseconds)
            wrote_nanoseconds = True
            break

    if not wrote_seconds:
        raise Exception(f"Destination {dest_object} does not have field encoding seconds")
    if not wrote_nanoseconds:
        raise Exception(f"Destination {source_object} does not have field encoding nanoseconds")

    return dest_object


def get_time_object(time_object):
    if "duration" in str(type(time_object)).lower():
        return fill_time(armw.Duration(), time_object)
    else:
        return fill_time(armw.Time(), time_object)

def get_native_time_object(armw_time: armw.Time | 'builtin_interfaces.msg.Time' | 'rospy.Time') -> 'builtin_interfaces.msg.Time' | 'rospy.Time':
    return fill_time(NativeTime(), armw_time)


from enum import IntEnum


class DocIntEnum(IntEnum):
    def __new__(cls, value, doc=None):
        self = int.__new__(cls, value)  # calling super().__new__(value) here would fail
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        else:
            self.__doc__ = ''
        return self

    @property
    def info(self):
        return f'{self.__class__}:{self.name}:[{self.value}]: {self.__doc__}'

    @classmethod
    def _get_field_index(cls, key: int):
        try:
            member_list = list(cls.__members__.items())
            for i in range(len(member_list)):  # Have to search through, because we might not have sequential values
                member = member_list[i]
                value = int(member[1])
                if value == key:
                    return i
            return None
        except IndexError:
            return None

    @classmethod
    def get_field(cls, key: int):
        index = cls._get_field_index(key)

        if index is not None:
            return list(cls.__members__.items())[index][1]
        else:
            return None

    @classmethod
    def get_field_names(cls):
        fields = list(cls.__members__.items())
        return [str(field[0]) for field in fields]

    @classmethod
    def get_field_name(cls, key: int):
        index = cls._get_field_index(key)

        if index is not None:
            return list(cls.__members__.items())[index][0]
        else:
            return None

    @classmethod
    def get_description(cls, key: int):
        """
        Attempts to return the best description of a field that it can
        """

        doc = cls.get_field(key).__doc__
        if len(doc) > 0:
            return doc
        else:
            return cls.get_field_name(key)



"""
Functions to handle importing interface definitions
"""

import importlib
from collections import OrderedDict
from typing import List, Dict

import roslib.message


def import_message(package_name: str, message_name: str):
    return getattr(importlib.import_module(f"{package_name}.msg"), message_name)


def import_service_(package_name: str, service_name: str):
    service_class = getattr(importlib.import_module(f"{package_name}.srv"), service_name)
    service_request = getattr(importlib.import_module(f"{package_name}.srv"), f"{service_name}Request")
    service_response = getattr(importlib.import_module(f"{package_name}.srv"), f"{service_name}Response")

    return service_class, service_response, service_request


def deserialze_anymsg(msg_data):
    topic_type = msg_data._connection_header['type']
    topic_class = roslib.message.get_message_class(topic_type)
    msg = topic_class()
    msg.deserialize(msg_data._buff)
    return msg


def get_msg_type_from_object(msg) -> str:
    return msg._type


def get_msg_fields_and_types(msg) -> Dict[str, str]:
    slot_names = msg.__slots__
    slot_types = msg._slot_types

    return OrderedDict(zip(slot_names, slot_types))


def get_msg_fields(msg) -> List[str]:
    return msg.__slots__



"""
Base node class that contains all the interfaces needed for the abstract middleware
"""

import inspect

import armw.globals


class BaseNode(object):
    def __init__(self):
        # Set the global node object before we run the rest of the constructor
        armw.globals.NODE = self

        self.update_rate = 1
        self.stop_commanded = False

    def spin(self):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def loop_once(self):
        pass

    def stop_running(self):
        self.stop_commanded = True

    def shutdown(self):
        pass

    def log(self, level, message):
        """
        Abstract function to log to console
        :param level: 0 for info, 1 for warn, 2 for error
        :param message: String to log
        :return: None
        """
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def log_debug(self, message, *args):
        self.log(-1, message % args)

    def log_info(self, message, *args):
        self.log(0, message % args)

    def log_warn(self, message, *args):
        self.log(1, message % args)

    def log_error(self, message, *args):
        self.log(2, message % args)

    def log_fatal(self, message, *args):
        self.log(3, message % args)

    def log_throttle(self, interval: int, level: int, message: str):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def log_debug_throttle(self, interval, message, *args):
        self.log_throttle(interval, -1, message % args)

    def log_info_throttle(self, interval, message, *args):
        self.log_throttle(interval, 0, message % args)

    def log_warn_throttle(self, interval, message, *args):
        self.log_throttle(interval, 1, message % args)

    def log_error_throttle(self, interval, message, *args):
        self.log_throttle(interval, 2, message % args)

    def get_name(self):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def get_namespace(self):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def search_param(self, parameter_name):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def has_param(self, parameter_name):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def get_param(self, parameter_name, default=None):
        """
        Returns the value for a given parameter name, or None if it can't find it
        """
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def set_param(self, parameter_name, value):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def delete_param(self, parameter_name):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def get_topic_type(self, topic):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def publish(self, topic: str, message_type, queue_size=1, latch=False):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def subscribe(self, topic: str, message_type, callback, queue_size=1):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def create_service_server(self, name: str, service_type, callback):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def create_service_client(self, name: str, service_type, persistent=False):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def wait_for_message(self, topic, message_type, timeout=None):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")

    def wait_for_service(self, name, timeout=None):
        raise NotImplementedError(f"armw.BaseNode.{inspect.stack()[0][3]}() not implemented")



"""
Various functions to explore the ROS world
"""

import armw.globals
from armw.ros2.util import canonicalize_name, ns_join
from ament_index_python.packages import get_package_share_directory


def check_master():
    return True


def get_topic_list():
    names_types = armw.globals.NODE.get_topic_names_and_types()
    return [x[0] for x in names_types]


def get_node_list():
    node_namespaces = armw.globals.NODE.get_node_names_and_namespaces()
    return [canonicalize_name(ns_join(x[1], x[0])) for x in node_namespaces]


def get_package_path(package_name: str):
    return get_package_share_directory(package_name)



import armw

SEP = '/'


def namespace(*args):
    return armw.NODE().get_namespace()


def ns_join(part_1, part_2):
    return f"{part_1}{SEP}{part_2}"


def is_private(name):
    return name.startswith("~")


def canonicalize_name(name):
    """
    Put name in canonical form. Double slashes '//' are removed and
    name is returned without any trailing slash, e.g. /foo/bar
    @param name: ROS name
    @type  name: str
    """

    # Straight up copied from here: https://docs.ros.org/en/melodic/api/rospy/html/rospy.names-pysrc.html

    if not name or name == SEP:
        return name
    elif name[0] == SEP:
        return '/' + '/'.join([x for x in name.split(SEP) if x])
    else:
        return '/'.join([x for x in name.split(SEP) if x])


def resolve_name(name):
    if not name:  # empty string resolves to namespace
        return namespace()

    name = str(name)  # enforce string conversion else struct.pack might raise UnicodeDecodeError (see #3998)

    name = canonicalize_name(name)

    if name[0] == SEP:  # global name
        resolved_name = name
    elif is_private(name):  # ~name
        resolved_name = ns_join(namespace(), name[1:])
    else:  # relative
        resolved_name = ns_join(namespace(), name)

    # Just in case, because I'm pretty sure we might add double slashes by accident
    return canonicalize_name(resolved_name)



import inspect
from typing import List

import rclpy
from rcl_interfaces.msg import SetParametersResult

from armw.ros2.ros2_node import Ros2Node
from armw.doc_int_enum import DocIntEnum
from armw.base_parameter_server import BaseParameterServer


class Ros2ParameterServer(BaseParameterServer):
    def __init__(self, node_object: Ros2Node, namespace: str = ""):
        super().__init__(node_object, namespace)
        self.node = node_object

        self.parameter_values = {}

        self.callback = None

    def add_parameter(self, name, description, default_value, min_value=None, max_value=None, parameter_type=None):
        if inspect.isclass(parameter_type) and issubclass(parameter_type, DocIntEnum):
            # TODO: no idea yet
            pass

        if not self.node.has_parameter(name):
            self.node.declare_parameter(name, default_value)

        self.parameter_values[name] = self.node.get_parameter(name).value

    def start(self, callback):
        self.node.add_on_set_parameters_callback(self.parameter_callback)

        # Save what callback we need to run when a parameter is changed
        self.callback = callback

        # Run the callback with all the startup values so we ensure that they get set to whatever we just got from the parameter server
        self.parameter_callback([])

    def parameter_callback(self, params: List[rclpy.Parameter]):
        for param in params:
            self.parameter_values[param.name] = param.value

        if self.callback is not None:
            self.callback(self.parameter_values, {})
            return SetParametersResult(successful=True)
        else:
            return SetParametersResult(successful=False)
